//---------------------------------------------------------------------------
#include "Precompiled.h"
#pragma hdrstop

#include <babel.h>
#include <gimite/temp_array.h>
#include "Win32Common.h"
#include "File.h"
using namespace std;
using namespace gimite;
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
istream& operator>>(istream& is, TCharCode& ccode)
{
	string s;
	if (is>>s){
		if (s=="Auto") ccode= ccAuto;
		else if (s=="SJIS") ccode= ccSjis;
		else if (s=="UTF16") ccode= ccUtf16;
		else if (s=="ASCII") ccode= ccAscii;
		else if (s=="EUC") ccode= ccEuc;
		else if (s=="JIS") ccode= ccJis;
		else if (s=="UTF8") ccode= ccUtf8;
		else if (s=="UTF8N") ccode= ccUtf8N;
		else if (s=="Unknown") ccode= ccUnknown;
		else is.setstate(ios::badbit);
	}
	return is;
}

ostream& operator<<(ostream& os, TCharCode ccode)
{
	switch (ccode){
		case ccSjis: os<<"SJIS"; break;
		case ccUtf16: os<<"UTF16"; break;
		case ccAscii: os<<"ASCII"; break;
		case ccEuc: os<<"EUC"; break;
		case ccJis: os<<"JIS"; break;
		case ccUtf8: os<<"UTF8"; break;
		case ccUtf8N: os<<"UTF8N"; break;
		case ccUnknown: os<<"Unknown"; break;
		case ccAuto: os<<"Auto"; break;
		default: assert(false);
	}
	return os;
}

//---------------------------------------------------------------------------
TFile<char>::TResult TFile<char>::LoadText(const TInfo& info/*= TInfo()*/)
	//eLXg[h
{
	TResult r= LoadTextSimply(Text(), info);
	if (r==forOk) DecodeReturnCode(Text());
	return r;
}

//---------------------------------------------------------------------------
TFile<char>::TResult TFile<char>::SaveText(const TInfo& info/*= TInfo()*/)
	//eLXgZ[u
{
	EncodeReturnCode(Text());
	return SaveTextSimply(Text());
}

//---------------------------------------------------------------------------
TFile<wchar_t>::TResult TFile<wchar_t>::LoadText(const TInfo& info)
	//eLXg[h
{
	GIMITE_MAPROFILED("TFile<wchar_t>::LoadText");
	assert(ccode_!=ccAscii);
	babel::init_babel();
	string stext;
	TResult r= LoadTextSimply(stext, info);
	if (r!=forOk) return r;
	if (ccode_==ccAuto) ccode_= AnalyzeCharCode(stext);
	if (ccode_==ccUnknown || ccode_==ccAscii) ccode_= ccSjis;
		//sASCIISJISƂĈ
	if (!DecodeCharCode(stext, Text())){
		if (r==forOk) r= forCharCodeMismatch;
	}
	if (r==forOk && info.bCheckCharCode_){
		if (!IsWideCharCode(ccode_) && stext.find('\0')!=string::npos){
			r= forBinary;//oCit@C
		}else{
			string stext2;
			EncodeCharCode(Text(), stext2);
			if (stext!=stext2) r= forCharCodeMismatch;//R[hsv
		}
	}
	DecodeReturnCode(Text());
	return r;
}

//---------------------------------------------------------------------------
TFile<wchar_t>::TResult TFile<wchar_t>::SaveText(const TInfo& info)
	//eLXgZ[u
{
	GIMITE_MAPROFILED("TFile<wchar_t>::SaveText");
	TString text= Text();
	EncodeReturnCode(text);
	string stext;
	EncodeCharCode(text, stext);
	return SaveTextSimply(stext);
}

//---------------------------------------------------------------------------
TCharCode TFile<wchar_t>::AnalyzeCharCode(const std::string& text)
	//R[h𔻒
{
	GIMITE_MAPROFILED("TFile<wchar_t>::AnalyzeCharCode");
	if (text.length()>=2 && text.substr(0, 2)=="\xff\xfe")
		return ccUtf16;
	if (text.length()>=3 && text.substr(0, 3)=="\xef\xbb\xbf")
		return ccUtf8;
	string::size_type pos;
	for (pos= 0; pos<text.length(); ++pos){
		if (!isprint(text[pos]) && text[pos]!='\n' && text[pos]!='\r'
				&& text[pos]!='\t')
			break;//ASCIIoĂ܂ŃXLbv
	}
	pos= pos/4*4;//Unicoden̔̂߂4̔{
	int enc= babel::analyze_base_encoding(text.substr(pos), 2000);
	switch (enc){
		case babel::base_encoding::sjis:
			return ccSjis;
		case babel::base_encoding::jis:
			return ccJis;
		case babel::base_encoding::euc:
			return ccEuc;
		case babel::base_encoding::iso2022jp:
			return ccJis;
		case babel::base_encoding::utf8:
			return ccUtf8N;
		case babel::base_encoding::ansi:
			return ccAscii;
		default:
			return ccUnknown;
	}
}

//---------------------------------------------------------------------------
bool TFile<wchar_t>::DecodeCharCode(const std::string& sSrc, TString& tsDest)
	//R[hUnicodeɕϊ
{
	GIMITE_MAPROFILED("TFile<wchar_t>::DecodeCharCode");
	babel::init_babel();
	bool bOk= true;
	switch (ccode_){
		case ccSjis:
			tsDest= SjisToUtf16(sSrc);
			break;
		case ccEuc:
			tsDest= babel::euc_to_unicode(sSrc);
			break;
		case ccJis:
			tsDest= babel::jis_to_unicode(sSrc);
			break;
		case ccUtf8:
			if (sSrc.length()>=3 && sSrc.substr(0, 3)=="\xef\xbb\xbf"){
				tsDest= babel::utf8_to_unicode(sSrc.substr(3));
			}else{
				tsDest= babel::utf8_to_unicode(sSrc);
				bOk= false;
			}
			break;
		case ccUtf8N:
			tsDest= babel::utf8_to_unicode(sSrc);
			break;
		case ccUtf16:{
			int lenBom;
			if (sSrc.length()>=2 && sSrc.substr(0, 2)=="\xff\xfe"){
				lenBom= 2;
			}else{
				lenBom= 0;
				bOk= false;
			}
			size_t len= sSrc.length()-lenBom;
			temp_array<char> buf(len+3);
			memcpy(buf.get(), sSrc.c_str()+lenBom, len+1);
			buf[len+1]= buf[len+2]= '\0';
			tsDest= reinterpret_cast<wchar_t*>(buf.get());
			break;
		}
		default:
			assert(false);
	}
	return bOk;
}

//---------------------------------------------------------------------------
void TFile<wchar_t>::EncodeCharCode(const TString& tsSrc, std::string& sDest)
	//R[hɖ߂
{
	GIMITE_MAPROFILED("TFile<wchar_t>::EncodeCharCode");
	babel::init_babel();
	switch (ccode_){
		case ccSjis:
			sDest= babel::unicode_to_sjis(tsSrc);
			break;
		case ccEuc:
			sDest= babel::unicode_to_euc(tsSrc);
			break;
		case ccJis:
			sDest= babel::unicode_to_jis(tsSrc);
			break;
		case ccUtf8:
			sDest= "\xef\xbb\xbf"+babel::unicode_to_utf8(tsSrc);
			break;
		case ccUtf8N:
			sDest= babel::unicode_to_utf8(tsSrc);
			break;
		case ccUtf16:
			sDest= "\xff\xfe"+string((char*)tsSrc.c_str(), tsSrc.length()*2);
			break;
		default:
			assert(false);
	};
}

//---------------------------------------------------------------------------

