//---------------------------------------------------------------------------
#ifndef PlaceH
#define PlaceH
//---------------------------------------------------------------------------
#include <string>
#include <vector>
#include <algorithm>
#include <babel.h>
//---------------------------------------------------------------------------
//ʒűʂ̊NX
class TAbstractPlace
{
public:
	TAbstractPlace(): bSelected_(true){}
	virtual ~TAbstractPlace(){}
	std::string Path()const{return path_;}
  int LineNumber()const{return nLine_;}
  int Position()const{return pos_;}
  int Length()const{return len_;}
  bool Selected()const{return bSelected_;}
  void SetPath(const std::string& path){path_= path;}
  void SetSelected(bool b){bSelected_= b;}
  virtual std::string Str()const= 0;//Y
  virtual std::string PreviousStr()const= 0;//O̕
  virtual std::string NextStr()const= 0;//̕

protected:
	void SetLineNumber(int n){nLine_= n;}
  void SetPosition(int pos){pos_= pos;}
  void SetLength(int len){len_= len;}

private:
	std::string path_;//pX
  int nLine_;//sڂ
  int pos_;//ڂ
  int len_;//
  bool bSelected_;//I𒆂

};

//---------------------------------------------------------------------------
//ʒůNX
template <class TChar>
class TPlaceBase: public TAbstractPlace
{
public:
	typedef typename std::basic_string<TChar> TString;
	typedef std::vector<TString> TStringVector;
	typedef std::vector<typename TString::size_type> TSizes;

public:
	TPlaceBase(const std::string& path, int nLine, const TString& text,
		int pos, int len, const TStringVector& ssRef,
    const TSizes& posesReturn);

	TString Tstr()const{return ts_;}
  TString PreviousTstr()const{return tsPrev_;}
  TString NextTstr()const{return tsNext_;}
  int RefCount()const{return tssRef_.size();}
  TString RefTstr(int i)const;

private:
  static const int lenAroundMax= 100;//ۑO̍̕ő咷

  TString ts_;//Y
  TString tsPrev_;//O̕
  TString tsNext_;//̕
	TStringVector tssRef_;//OQƕ

};

//---------------------------------------------------------------------------
template <class TChar>
class TPlace;
//---------------------------------------------------------------------------
//1oCg̈ʒu
template <>
class TPlace<char>: public TPlaceBase<char>
{
public:
	typedef TPlaceBase<char>::TString TString;
	typedef TPlaceBase<char>::TStringVector TStringVector;
	typedef TPlaceBase<char>::TSizes TSizes;

	TPlace(const std::string& path, int nLine, const TString& text,
		int pos, int len, const TStringVector& ssRef, const TSizes& posesReturn)
		: TPlaceBase<char>(path, nLine, text, pos, len, ssRef, posesReturn){}
	virtual std::string Str()const{return Tstr();}//Y
	virtual std::string PreviousStr()const{return PreviousTstr();}
		//O̕
	virtual std::string NextStr()const{return NextTstr();}//̕

};

//---------------------------------------------------------------------------
//2oCg̈ʒu
template <>
class TPlace<wchar_t>: public TPlaceBase<wchar_t>
{
public:
  typedef TPlaceBase<wchar_t>::TString TString;
  typedef TPlaceBase<wchar_t>::TStringVector TStringVector;

	TPlace(const std::string& path, int nLine, const TString& text,
		int pos, int len, const TStringVector& ssRef, const TSizes& posesReturn);
  virtual std::string Str()const//Y
  	{return babel::unicode_to_sjis(Tstr());}
  virtual std::string PreviousStr()const//O̕
  	{return babel::unicode_to_sjis(PreviousTstr());}
  virtual std::string NextStr()const//̕
  	{return babel::unicode_to_sjis(NextTstr());}

};

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//ev[g֐̎
//---------------------------------------------------------------------------
template <class T>
TPlaceBase<T>::TPlaceBase(const std::string& path, int nLine,
	const TString& text, int pos, int len, const TStringVector& tssRef,
  const TSizes& posesReturn)
{
  SetPath(path);
  std::size_t nRelLine=
	  	std::lower_bound(posesReturn.begin(), posesReturn.end(), pos)
      -posesReturn.begin();
  	//sʒu^Ă΁Agtextŉsڂ߂
  SetLineNumber(nLine+nRelLine);
  SetPosition(pos-(nRelLine==0? 0:(posesReturn[nRelLine-1]+1)));
  SetLength(len);
  ts_= text.substr(pos, len);
  TString::size_type posPrev= std::max(0, pos-lenAroundMax);
  tsPrev_= text.substr(posPrev, pos-posPrev);
  tsNext_= text.substr(pos+len,
  	std::min(lenAroundMax, text.length()-len-pos));
  tssRef_= tssRef;
}

//---------------------------------------------------------------------------
template <class T>
TPlaceBase<T>::TString TPlaceBase<T>::RefTstr(int i)const
{
  if (i>=0 && i<tssRef_.size()) return tssRef_[i];
  else return TString();
}

//---------------------------------------------------------------------------
#endif
